import glob
import yaml

from jinja2 import Environment, BaseLoader
from rich.console import Console
from rich.table import Table
from os.path import abspath, dirname, join

console = Console()


def update_namespace(namespace, new_file, verbose):

    if namespace is None:
        return new_file
    for key, value in new_file.items():
        if key in namespace and key in ("checks", "datasources", "connections"):
            namespace[key] = namespace[key] + new_file[key]
        elif key in namespace and key in ("includes"):  # remove duplicates
            namespace[key] = list(set(namespace[key] + new_file[key]))
        elif key in ("checks", "datasources", "includes", "connections"):
            namespace[key] = new_file[key]
        elif key in ("extras"):
            pass  # ignored keys
        elif verbose:
            console.print(f"Key not supported yet: {key}")
    return namespace


def load_config(
    config_path: str,
    namespace: dict = None,
    context: dict = None,
    visited_path: dict = None,
    table: Table = None,
    verbose: bool = True,
    first_run: bool = True,
) -> dict:
    print_final_table = False
    if first_run:
        first_run = False
        if verbose:
            print_final_table = True
            table = Table("File Path", "# of checks")
        visited_path = {}

    file_paths = glob.glob(config_path)
    if verbose:
        console.print(f"Walking Paths: {file_paths}")
    for file_path in file_paths:
        if file_path in visited_path:
            continue
        visited_path[file_path] = True
        with open(file_path, "r") as stream:
            if context:
                data_loaded = yaml.safe_load(
                    Environment(loader=BaseLoader())
                    .from_string(stream.read())
                    .render(context)
                )
            else:
                data_loaded = yaml.safe_load(stream)
            if verbose:
                table.add_row(str(file_path), str(len(data_loaded["checks"])))

        if "includes" in data_loaded:
            for included_path in data_loaded["includes"]:
                if (
                    namespace
                    and "includes" in namespace
                    and included_path in namespace["includes"]
                ):
                    continue
                if included_path.startswith("/"):
                    included_path = included_path[1:]
                root_dir = dirname(abspath(file_path))
                namespace = load_config(
                    join(root_dir, included_path),
                    namespace=namespace,
                    visited_path=visited_path,
                    table=table,
                    verbose=verbose,
                    first_run=first_run,
                )
        namespace = update_namespace(namespace, data_loaded, verbose)

    if print_final_table:
        console.print(table)

    return namespace



import os
import typer
import datetime

from dotenv import load_dotenv
from rich import print
from typing_extensions import Annotated


from weiser.loader.export import export_results, print_results
from weiser.loader.config import load_config
from weiser.runner import pre_run_config, run_checks, generate_sample_data


# Initialize Typer
app = typer.Typer()
version = "0.1.9"


@app.callback()
def callback():
    """
    Weiser is a data quality framework designed to help you ensure the integrity and accuracy of your data.
    It provides a set of tools and checks to validate your data and detect anomalies.
    It also includes a dashboard to visualize the results of the checks.
    """
    print(
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] [bold red]Running Weiser version:[/bold red] [green]{version}[/green] :boom:\n"
    )


@app.command()
def run(
    input_config: Annotated[str, typer.Argument(help="The path for the file to read")],
    verbose: Annotated[
        bool, typer.Option("--verbose", "-v", help="Print to stdout the parsed files")
    ] = False,
    show_ids: Annotated[
        bool, typer.Option("--show-ids", "-i", help="Print check ids to results table")
    ] = False,
    skip_export: Annotated[
        bool, typer.Option("--skip-export", "-s", help="Skip exporting results")
    ] = False,
):
    """
    Main Command
    """
    # Load .env
    if os.path.exists(".env"):
        if verbose:
            print("Loading .env file")
        load_dotenv(
            dotenv_path=".env",
            verbose=verbose,
        )
    env_variables = dict(os.environ)
    config = load_config(input_config, context=env_variables)
    context = pre_run_config(config, verbose=verbose)
    results = run_checks(
        context["run_id"],
        context["config"],
        context["connections"],
        context["metric_store"],
        verbose,
    )
    if not skip_export:
        export_results(context["run_id"], context["metric_store"])
    print_results(results, show_ids)
    print(
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] [green]Finished Run[/green] :rocket:"
    )


@app.command()
def compile(
    input_config: Annotated[str, typer.Argument(help="The path for the file to read")],
    verbose: Annotated[
        bool, typer.Option("--verbose", "-v", help="Print to stdout the parsed files")
    ] = False,
):
    """
    Main Command
    """
    # Load .env
    if os.path.exists(".env"):
        load_dotenv(
            dotenv_path=".env",
            verbose=verbose,
        )
    config = load_config(input_config)
    pre_run_config(config, compile_only=True, verbose=verbose)
    print(
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] [green]Finished Config compilation[/green] :rocket:"
    )


@app.command()
def sample(
    input_config: Annotated[str, typer.Argument(help="The path for the file to read")],
    check: Annotated[str, typer.Option("--check", "-c", help="Id to populate")],
    verbose: Annotated[
        bool, typer.Option("--verbose", "-v", help="Print to stdout the parsed files")
    ] = False,
    skip_export: Annotated[
        bool, typer.Option("--skip-export", "-s", help="Skip exporting results")
    ] = False,
):
    """
    Generate sample data based on a check id name.
    """
    # Load .env
    if os.path.exists(".env"):
        load_dotenv(
            dotenv_path=".env",
            verbose=verbose,
        )
    config = load_config(input_config)
    context = pre_run_config(config, verbose)
    results = generate_sample_data(
        check,
        context["config"],
        context["connections"],
        context["metric_store"],
        verbose,
    )
    if not skip_export:
        export_results(context["run_id"], config)
    print(
        f"[{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}] [green]Finished Generating Sample[/green] :rocket:"
    )


if __name__ == "__main__":
    app()



from rich.console import Console
from rich.table import Table, Column
from weiser.drivers.metric_stores import MetricStoreDB

console = Console()


def export_results(run_id: str, metric_store: MetricStoreDB):
    metric_store.export_results(run_id)
    return True


def print_results(results, show_ids: bool):
    columns = [
        "Check Name",
        "Datasource",
        "Dataset",
        "Measure",
        "Condition",
        "Actual Value",
        "Threshold",
        "Result",
    ]
    if show_ids:
        # columns = [Column(header="Check Id", overflow="fold")] + columns
        columns = ["Check Id"] + columns
    table = Table(*columns)
    for results_item in results:
        for result in results_item["results"]:
            row = [
                result.get("name"),
                result.get("datasource"),
                result.get("dataset"),
                result.get("measure") or result.get("type"),
                result.get("condition"),
                str(result.get("actual_value")),
                str(
                    result.get("threshold_list")
                    if result.get("threshold") is None
                    else result.get("threshold")
                ),
                ":x:" if result.get("fail") else ":white_check_mark:",
            ]
            if show_ids:
                row = [result.get("check_id")] + row
            table.add_row(*row)
    console.print(table)



from typing import Union
from weiser.drivers.metric_stores.duckdb import DuckDBMetricStore
from weiser.drivers.metric_stores.postgres import PostgresMetricStore
from weiser.loader.models import MetricStore, MetricStoreType

QUERY_TYPE_MAP = {
    MetricStoreType.duckdb: DuckDBMetricStore,
    MetricStoreType.postgresql: PostgresMetricStore,
}

MetricStoreDB = Union[DuckDBMetricStore, PostgresMetricStore]
# MetricStoreDB = DuckDBMetricStore


class MetricStoreFactory:
    @staticmethod
    def create_driver(metric_store: MetricStore) -> MetricStoreDB:
        if metric_store:
            return QUERY_TYPE_MAP.get(
                metric_store.db_type,
                DuckDBMetricStore,
            )(metric_store)
        return DuckDBMetricStore(MetricStore())



from sqlglot.expressions import Select
from weiser.checks.base import BaseCheck
from weiser.loader.models import DBType


class CheckNumeric(BaseCheck):
    def get_query(self, table: str, verbose: bool) -> Select:
        return self.build_query(
            [
                self.check.measure,
            ],
            table,
            verbose=verbose,
        )


# Cube specific check for Measures
class CheckMeasure(BaseCheck):
    def get_query(self, table: str, verbose: bool) -> Select:
        return self.build_query(
            [
                f"MEASURE({self.check.measure})",
            ],
            table,
            verbose=verbose,
        )


class CheckRowCount(BaseCheck):
    def get_query(self, table: str, verbose: bool) -> Select:
        return self.build_query(
            [
                "COUNT(*)",
            ],
            table,
            verbose=verbose,
        )


class CheckSum(BaseCheck):
    def get_query(self, table: str, verbose: bool) -> Select:
        return self.build_query(
            [
                f"SUM({self.check.measure})",
            ],
            table,
            verbose=verbose,
        )


class CheckMax(BaseCheck):
    def get_query(self, table: str, verbose: bool) -> Select:
        return self.build_query(
            [
                f"MAX({self.check.measure})",
            ],
            table,
            verbose=verbose,
        )


class CheckMin(BaseCheck):
    def get_query(self, table: str, verbose: bool) -> Select:
        return self.build_query(
            [
                f"MIN({self.check.measure})",
            ],
            table,
            verbose=verbose,
        )



from pprint import pprint
from typing import Any
from sqlalchemy import create_engine, text
from sqlalchemy.engine import URL
from sqlglot.expressions import insert, values, Select
from sqlglot.dialects import Postgres
from typing import List, Tuple

from weiser.loader.models import MetricStore


class PostgresMetricStore:
    def __init__(self, metric_store: MetricStore) -> None:
        self.config = metric_store
        if not metric_store.uri:
            uri = URL.create(
                metric_store.db_type,
                username=metric_store.user,
                password=metric_store.password.get_secret_value(),
                host=metric_store.host,
                database=metric_store.db_name,
            )
        else:
            uri = metric_store.uri

        self.engine = create_engine(uri)
        self.db_name = metric_store.db_name
        self.dialect = Postgres

        with self.engine.connect() as conn:
            conn.execute(
                text(
                    """CREATE TABLE IF NOT EXISTS metrics (
                            actual_value double precision,
                            check_id VARCHAR,
                            condition VARCHAR,
                            dataset VARCHAR,
                            datasource VARCHAR,
                            fail BOOLEAN,
                            name VARCHAR,
                            run_id VARCHAR,
                            run_time TIMESTAMP,
                            sql VARCHAR,
                            success boolean,
                            threshold VARCHAR,
                            threshold_list double precision[],
                            type VARCHAR
                            )"""
                )
            )

    # Meant for metadata queries, like anomaly detection
    def execute_query(
        self,
        q: Select,
        check: Any,
        verbose: bool = False,
        validate_results: bool = True,
    ):
        engine = self.engine
        with engine.connect() as conn:
            rows = list(conn.execute(text(q.sql(dialect=self.dialect))))
            if (
                validate_results
                and not len(rows) > 0
                and not len(rows[0]) > 0
                and not rows[0][0] is None
            ):
                raise Exception(
                    f"Unexpected result executing check: {check.model_dump()}"
                )
            if verbose:
                pass
                # pprint(rows)
        return rows

    def insert_results(self, record):
        with self.engine.connect() as conn:
            if isinstance(record["threshold"], List) or isinstance(
                record["threshold"], Tuple
            ):
                record["threshold_list"] = record["threshold"]
                record["threshold"] = None
            elif "threshold_list" not in record:
                record["threshold_list"] = None
            q = insert(
                values(
                    [
                        (
                            record["actual_value"],
                            record["check_id"],
                            record["condition"],
                            record["dataset"],
                            record["datasource"],
                            record["fail"],
                            record["name"],
                            record["run_id"],
                            record["run_time"],
                            record["measure"],
                            record["success"],
                            record["threshold"],
                            record["threshold_list"],
                            record["type"],
                        )
                    ]
                ),
                "metrics",
            )
            conn.execute(text(q.sql(dialect="postgres")))

    def export_results(self, run_id):
        pass



import boto3
import duckdb
from rich import print
from typing import List, Tuple, Any

from sqlglot.expressions import insert, values, Select
from sqlglot.dialects import DuckDB
from weiser.loader.models import MetricStore, S3UrlStyle


class DuckDBMetricStore:
    def __init__(self, config: MetricStore) -> None:
        self.config = config
        self.s3_client = boto3.client("s3")
        self.db_name = config.db_name
        self.dialect = DuckDB
        if not self.db_name:
            self.db_name = "./metricstore.db"
        with duckdb.connect(self.db_name) as conn:
            conn.sql("INSTALL httpfs;")
            conn.sql("LOAD httpfs;")
            if self.config.s3_url_style == S3UrlStyle.path:
                conn.sql(f"SET s3_url_style='{self.config.s3_url_style}'")
            elif self.config.s3_url_style == S3UrlStyle.vhost:
                conn.sql(f"SET s3_region = '{self.config.s3_region}'")
            if self.config.s3_endpoint:
                conn.sql(f"SET s3_endpoint = '{self.config.s3_endpoint}'")
            conn.sql(f"SET s3_access_key_id = '{self.config.s3_access_key}'")
            conn.sql(f"SET s3_secret_access_key = '{self.config.s3_secret_access_key}'")
            conn.sql(
                """CREATE TABLE IF NOT EXISTS metrics (
                     actual_value DOUBLE,
                     check_id VARCHAR,
                     condition VARCHAR,
                     dataset VARCHAR,
                     datasource VARCHAR,
                     fail BOOLEAN,
                     name VARCHAR,
                     run_id VARCHAR,
                     run_time TIMESTAMP,
                     sql VARCHAR,
                     success boolean,
                     threshold VARCHAR,
                     threshold_list DOUBLE[],
                     type VARCHAR
                     )"""
            )
            res = conn.sql(
                """SELECT MAX(run_time) AS run_time FROM metrics"""
            ).fetchall()
            last_run_time = "1=1"
            if res and res[0][0]:
                last_run_time = f"run_time > '{res[0][0]}'"
            conn.sql(
                f"""
                INSERT INTO metrics SELECT * FROM 's3://{self.config.s3_bucket}/metrics/*.parquet' WHERE {last_run_time};
                """
            )

    # Delete Parquet files
    def delete_parquet_files(self, prefix):
        bucket_name = self.config.s3_bucket
        response = self.s3_client.list_objects_v2(Bucket=bucket_name, Prefix=prefix)

        if "Contents" in response:
            for obj in response["Contents"]:
                if obj["Key"].endswith(".parquet"):
                    self.s3_client.delete_object(Bucket=bucket_name, Key=obj["Key"])
                    # print(f"Deleted {obj['Key']}")

    # Move Parquet files
    def move_file(self, source_key, destination_key):
        bucket_name = self.config.s3_bucket
        copy_source = {"Bucket": bucket_name, "Key": source_key}
        # Copy the file to the new location
        self.s3_client.copy_object(
            CopySource=copy_source, Bucket=bucket_name, Key=destination_key
        )
        # Delete the original file
        self.s3_client.delete_object(Bucket=bucket_name, Key=source_key)

    # Meant for metadata queries, like anomaly detection
    def execute_query(
        self,
        q: Select,
        check: Any,
        verbose: bool = False,
        validate_results: bool = True,
    ):
        with duckdb.connect(self.db_name) as conn:
            rows = conn.sql(q.sql(dialect=self.dialect)).fetchall()
            if validate_results and not len(rows) > 0:
                if verbose:
                    print(q.sql(dialect=self.dialect))
                raise Exception(
                    f"Unexpected result executing check: {check.model_dump()}"
                )
            if verbose:
                # pprint(rows)
                pass
        return rows

    def insert_results(self, record):
        with duckdb.connect(self.db_name) as conn:
            if isinstance(record["threshold"], List) or isinstance(
                record["threshold"], Tuple
            ):
                record["threshold_list"] = record["threshold"]
                record["threshold"] = None
            elif "threshold_list" not in record:
                record["threshold_list"] = None
            q = insert(
                values(
                    [
                        (
                            record["actual_value"],
                            record["check_id"],
                            record["condition"],
                            record["dataset"],
                            record["datasource"],
                            record["fail"],
                            record["name"],
                            record["run_id"],
                            record["run_time"],
                            record["measure"],
                            record["success"],
                            record["threshold"],
                            record["threshold_list"],
                            record["type"],
                        )
                    ]
                ),
                "metrics",
            )
            conn.sql(q.sql(dialect="duckdb"))

    def export_results(self, run_id):
        if (
            not self.config.s3_bucket
            or not self.config.s3_access_key
            or not self.config.s3_secret_access_key
        ):
            print("No S3 bucket configured, skipping export")
            return
        with duckdb.connect(self.db_name) as conn:
            conn.sql("INSTALL httpfs;")
            conn.sql("LOAD httpfs;")
            if self.config.s3_url_style == S3UrlStyle.path:
                conn.sql(f"SET s3_url_style='{self.config.s3_url_style}'")
            elif self.config.s3_url_style == S3UrlStyle.vhost:
                conn.sql(f"SET s3_region = '{self.config.s3_region}'")
            if self.config.s3_endpoint:
                conn.sql(f"SET s3_endpoint = '{self.config.s3_endpoint}'")
            conn.sql(f"SET s3_access_key_id = '{self.config.s3_access_key}'")
            conn.sql(f"SET s3_secret_access_key = '{self.config.s3_secret_access_key}'")
            conn.sql(
                f"COPY (SELECT * FROM metrics WHERE run_id='{run_id}') TO 's3://{self.config.s3_bucket}/metrics/{run_id}.parquet' (FORMAT 'parquet');"
            )
            conn.sql(
                f"""
                    COPY (SELECT * FROM 's3://{self.config.s3_bucket}/metrics/*.parquet') TO 's3://{self.config.s3_bucket}/tmp/merged_at_{run_id}.parquet' (FORMAT 'parquet');
                """
            )
            # Delete old Parquet files
            self.delete_parquet_files("metrics/")
            # Move the merged file to the final location
            self.move_file(
                source_key=f"tmp/merged_at_{run_id}.parquet",
                destination_key=f"metrics/{run_id}.parquet",
            )
